/*
 * Copyright (c) 1998 Qualcomm Incorporated. All rights reserved.
 * 
 * snprintf.c : 
 */

#include <string.h>
#include <stdarg.h>
#include <stdio.h>

int snprintf( char *s, size_t n, const char *format, ...);

/*
 * A snprintf minimal implementation built over sprintf.
 * Extracted from K&R 'C'.
 * Reference : Unix specification version 2. Copyright 1997 Open Group.
 * null terminated string.
 * Caveats :
 *    - Limited support for format modifiers.
 *    - Format modifiers not allowed for %s.
 *    - No nth argument specification %n$.
 *    - The known formats "sdcfxXgGeEou%".
 *    - Assuming int as 4 bytes, double as 8 bytes. No support for
 *      long double.
 *
 * Parameters : s - buffer to print to.
 *              n - max number of bytes to print.
 *              format - the format string.
 *
 * Returned : On success the number of bytes printed to buffer.
 *            On failure -1.
 */
#define BUFSIZE 1024

static void copy_buf( char **s, char *m, size_t *n )
{
    strncpy( *s, m, *n );
    *n     -= strlen(*s);
    *s     += strlen(*s);
}

int snprintf( char *s, size_t n, const char *format, ...)
{
    char    *sval;
    int     ival;
    double  dval;
    char    cval;
    float   fval;

    va_list  ap;                         /* Pointer into stack to extract
					  * parameters */
    char    *p = (char *)format,         /* Pointer into format string */
	     msgBuf[BUFSIZE],            /* Temporary store for printing
					  * ints, floats etc using sprintf */
	     frmToken[64],               /* Temporary store for conversion
					  * tokens. */
            *f,                          /* Pointer in to frmToken */
	     nState=0;                   /* 0 when copying the format, 
					  * 1 when reading the conversion
					  * specification. */
    size_t   nSize = n-1;

    if ( s == 0 || n == 0 || format == 0 ) return -1;
    s[n-1] = '\0';

    va_start(ap, format);
    for( ; nSize && *p; p++ ){
	if( nState == 0 ) {
	    if( *p != '%' ) {
		*s++ = *p;
		nSize--;
	    }
	    else {
		nState = 1;
		f = frmToken;
		*f++ = *p;
	    }
	}
	else {
	    *f++ = *p;
	    *f = '\0';
	    switch( *p ) {
	    case 'i':
	    case 'c':
	    case 'd':
	    case 'o':
	    case 'u':
	    case 'x':
	    case 'X':
		ival = va_arg( ap, int );    /* Args on stack are word aligned */
		if( frmToken[1] == '*' ) || ( isdigit( frmToken[1] ) && 
					      atoi( &frmToken[1] ) > 
					      min( BUFSIZE, nSize ) ) return -1;
		sprintf( msgBuf, frmToken, ival );
		copy_buf( &s, msgBuf, &nSize );
		nState = 0;
		break;
	    case 'f':
	    case 'e':
	    case 'E':
	    case 'g':
	    case 'G':
		if( frmToken[1] == '*' ) || ( isdigit( frmToken[1] ) && 
					      atoi( &frmToken[1] ) > 
					      min( BUFSIZE, nSize ) ) return -1;
		dval = va_arg(ap, double);
		sprintf(msgBuf, frmToken, dval);
		copy_buf( &s, msgBuf, &nSize );
		nState = 0;
		break;
	    case 's':
		if( frmToken[1] == '*' ) || ( isdigit( frmToken[1] ) && 
					      atoi( &frmToken[1] ) > 
					      min( BUFSIZE, nSize ) ) return -1;
		sval = va_arg(ap, char *);
		copy_buf( &s, sval, &nSize );
		nState = 0;
		break;
	    case '%':
		*s++ = *p;
		nSize--;
		nState = 0;
		break;
	    default:
		break;
	    }
	}
    }
    va_end(ap);
    if( nSize ) *s++ = *p;
    return ( (n-1) - nSize );
}
