/*
 * Copyright (c) 1989 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#include <config.h>
#include <stdio.h>
#include <sys/types.h>
#include <popper.h>
#include "mmangle/mime.h"
#include "mmangle/mangle.h"

void mangle_count(mstate, buf, len)
     void *mstate;
     char *buf;
     int   len;
{
  char *p;

  *(long *)mstate += len;

  for(p = buf; len > 0; p++, len--)
    if(*p == '\n')
      (*(long *)mstate)++;

}

/* 
 *  list:   List the contents of a POP maildrop
 */

int pop_list (p)
POP     *   p;
{
    MsgInfoList         *   mp;         /*  Pointer to message info list */
    register int            i;
    register int            msg_num;
    int                     show_mangled_length=0;
    unsigned int            mangled_length=0;
    char                    buffer[MAXMSGLINELEN];
    unsigned int            msg_lines;
    MimeParsePtr            mimeParse;
    ManglerStateType        mangleState;

    /*  Were arguments provided ? */
    pop_lower(p->pop_parm[1]);
    if(p->parm_count == 1 && strncmp(p->pop_parm[1], "mangle",6) == 0) {
	show_mangled_length++;
    } 
    else if(p->parm_count == 1 && strncmp(p->pop_parm[1], "x-mangle", 8) ==0) {
        show_mangled_length++;
    }
    else if(p->parm_count > 0) {	
        /* First arg must be a message number */
        msg_num = atoi(p->pop_parm[1]);


        /*  Is requested message out of range? */
        if ((msg_num < 1) || (msg_num > p->msg_count))
            return (pop_msg (p,POP_FAILURE,
			     "Message %d does not exist.",msg_num));

        /*  Get a pointer to the message in the message list */
        mp = &p->mlp[msg_num-1];

        /*  Is the message already flagged for deletion? */
        if (mp->del_flag)
            return (pop_msg (p,POP_FAILURE,
			     "Message %d has been deleted.",msg_num));

        /* If second arg, it must be "mangle" or "x-mangle" */
        if(p->parm_count == 2) {
	    pop_lower(p->pop_parm[2]);
	    if((strncmp(p->pop_parm[2], "mangle",6) != 0) &&
               (strncmp(p->pop_parm[2], "x-mangle", 8) != 0) )
		return(pop_msg(p, POP_FAILURE,"Unknown LIST argument: %s", 
			       p->pop_parm[2]));
	    else show_mangled_length++;
        }
	if(show_mangled_length){
	    (void) memset((void *)&mangleState, 0, sizeof(ManglerStateType));
	    mangleState.outFn = mangle_count;
	    mangleState.outFnState = &(mangled_length);
	    mangleState.lastWasGoodHeader = 0;
	    mangleState.outFnConstructor  = mangleState.outFnDestructor 
		= NULL;
	    if(FillMangleInfo(p->pop_parm[2],&mangleState.rqInfo) == -1) {
		return(pop_msg(p, POP_FAILURE,"Syntax error in x-mangle"));
	    }
	    mimeParse = MimeInit(MangleMapper, &mangleState, p->drop);
	    (void)fseek(p->drop, mp->offset, 0);
	    /* Read the from envelope */
	    fgets(buffer,MAXMSGLINELEN,p->drop);
	    MSG_LINES = mp->lines;
	    while(fgets(buffer,MAXMSGLINELEN, p->drop) && MSG_LINES--) {
		MimeInput(mimeParse, buffer, strlen(buffer));
	    }
	    FreeMangleInfo(&mangleState.rqInfo);
	    MimeFinish(mimeParse);
	}

        /*  Display message information */
        return (pop_msg(p,POP_SUCCESS,"%u %u",msg_num,
			(show_mangled_length ? mangled_length : mp->length)));
    }

    
    /*  Display the entire list of messages */
    if(FillMangleInfo(p->pop_parm[1],&mangleState.rqInfo) == -1) {
	return(pop_msg(p, POP_FAILURE,"Syntax error in Mangle"));
    }
    pop_msg(p,POP_SUCCESS,
        "%u messages (%u octets)",
            p->msg_count-p->msgs_deleted,p->drop_size-p->bytes_deleted);

    /*  Loop through the message information list.  Skip deleted messages */
    for (i = p->msg_count, mp = p->mlp; i > 0; i--, mp++) {
        if (mp->del_flag)
	  continue;
	if(show_mangled_length){
	    (void) memset((void *)&mangleState, 0, sizeof(ManglerStateType));
	    mangled_length = 0;
	    mangleState.outFn = mangle_count;
	    mangleState.outFnState = &(mangled_length);
	    mangleState.lastWasGoodHeader = 0;
	    mangleState.outFnConstructor  = mangleState.outFnDestructor 
		= NULL;
	    FillMangleInfo(p->pop_parm[1],&mangleState.rqInfo);
	    mimeParse = MimeInit(MangleMapper, &mangleState, p->drop);
	    (void)fseek(p->drop, mp->offset, 0);
	    /* Read the from envelope */
	    fgets(buffer,MAXMSGLINELEN,p->drop);
	    MSG_LINES = mp->lines;
	    while(fgets(buffer,MAXMSGLINELEN, p->drop) && MSG_LINES--) {
		MimeInput(mimeParse, buffer, strlen(buffer));
	    }
	    FreeMangleInfo(&mangleState.rqInfo);
	    MimeFinish(mimeParse);
	}
        (void)fprintf(p->output,"%u %u\r\n", mp->number,
                     (show_mangled_length ? mangled_length : mp->length));
    }

    /*  "." signals the end of a multi-line transmission */
    (void)fprintf(p->output,".\r\n");
    (void)fflush(p->output);

    return(POP_SUCCESS);
}
